/* -*- c-basic-offset: 4 -*- */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "net.h"

#define NONE 0 
#define PHARAO 1
#define DJHED  2
#define PYRAMID 3
#define OBELISK 4
#define OBELISK2 5

#define N_PIECE_TYPES 6

#define SILVER 0
#define RED 1

#define F_ROTATE 1
#define F_SPLIT 2

#define BOARD_WIDTH 10
#define BOARD_HEIGHT 8

#define E_ILLEGAL_MOVE -1
#define E_UNKNOWN_COMMAND -2
#define HUGE_SUCCESS 1

// return values for laser():
#define L_WALL 0
#define L_PIECE 1
#define SILVER_WINS 2
#define RED_WINS 3

#define MAX_MSG_LEN 50
typedef enum { MSG_MOVE, MSG_ROTATE } move_t;
typedef enum { R_CW, R_CCW } rot_t;
typedef enum { D_NW=0, D_N, D_NE, D_E, D_SE, D_S, D_SW, D_W } dir_t;

typedef struct {
    move_t type;
    union {
	struct {
	    int row, col;
	    dir_t dir;
	} move;
	struct {
	    int row, col;
	    rot_t dir;
	} rotate;
    } args;
} message_t;

struct pieceinfo {
    int num;
} pieceinfo[] = {
    { // NONE
	.num = 0 },
    { // PHARAO
	.num = 1 },
    { // DJHED
	.num = 2 },
    { // PYRAMID
	.num = 7 },
    { // OBELISK
	.num = 4 },
    { // OBELISK2
	.num = 0 },
};

struct square {
    unsigned int piece : 4;
    unsigned int side : 1;
    unsigned int dir : 2;
} board[BOARD_WIDTH][BOARD_HEIGHT];

int laser_pos[2];

struct sideinfo {
    char ch;
    char *name;
} sideinfo[] = {
    { .ch = 'o', .name = "silver" },
    { .ch = 'x', .name = "red" } };

char classic_setup[] =
    "H40" // PHARAO
    "E40E51" // DJHED
    "H22G73E22D21E91D92C32" // PYRAMID
    "H30H30H50H50" // OBELISK
    ;


char asciiart_small_pieces[] =
    "                " // NONE
    "SSSSSSSSRRRRRRRR"
    ",'`.,'`..\"\"..\"\"." // DJHED
    "`/\\'/.,\\\"/\\\"/==\\" // PYRAMID
    ". . . . = = = = " // OBELISK
    "........========" // OBELISK2
    ;

void
print_piece_small(int piece, int side, int dir)
{
    int index = (piece*4*2 + side*4 + dir) * 2;
    net_all_printf("%c%c",
		   asciiart_small_pieces[index],
		   asciiart_small_pieces[index+1]);
}

int
laser_at(int x, int y)
{
    return laser_pos[0] == x && laser_pos[1] == y;
}



// FIXME rewrite this to format the board in a char *?
void
print_board_small()
{
    int x, y;

    char vline, hline, xline, lline;
    char dead_piece[] = "##";
    vline = '|', hline = '-', xline = '+';
    lline = '#';

    net_all_printf("%s(%c)\n", sideinfo[RED].name, sideinfo[RED].ch);

    // column numbers
    net_all_printf("  %c", vline);
    for (x = 0; x < BOARD_WIDTH; x++) {
	net_all_printf(" %d%c", x, vline);
    }
    net_all_printf("\n");

    // line
    net_all_printf("%c%c%c", hline, hline, xline);
    for (x = 0; x < BOARD_WIDTH; x++)
	net_all_printf("%c%c%c",
	       laser_at(x, -1) ? lline : hline,
	       laser_at(x, -1) ? lline : hline,
	       xline);
    net_all_printf("%c%c\n", hline, hline);

    for (y = 0; y < BOARD_HEIGHT; y++) {
	// row name
	net_all_printf(" %c%c", 'A'+y, laser_at(-1,y) ? lline : vline);
	// row
	for (x = 0; x < BOARD_WIDTH; x++) {
	    if (laser_at(x, y))
		net_all_printf("%s", dead_piece);
	    else
		print_piece_small(board[x][y].piece,
				  board[x][y].side,
				  board[x][y].dir);
	    net_all_printf("%c", (x==BOARD_WIDTH-1 && laser_at(x+1, y)) ? lline : vline);
	}
	// row name
	net_all_printf("%c\n", 'A'+y);

	// line
	net_all_printf("%c%c%c", hline, hline, xline);
	for (x = 0; x < BOARD_WIDTH; x++)
	    net_all_printf("%c%c%c",
		   (y==BOARD_HEIGHT-1 && laser_at(x, y+1)) ? lline : hline,
		   (y==BOARD_HEIGHT-1 && laser_at(x, y+1)) ? lline : hline,
		   xline);
	net_all_printf("%c%c\n", hline, hline);
    }

    // column numbers
    net_all_printf("  %c", vline);
    for (x = 0; x < BOARD_WIDTH; x++)
	net_all_printf(" %d%c", x, vline);
    net_all_printf("\n");

    for (x = 0; x < (BOARD_WIDTH+2)*3-1-strlen(sideinfo[SILVER].name)-3; x++)
	net_all_printf(" ");
    net_all_printf("%s(%c)\n", sideinfo[SILVER].name, sideinfo[SILVER].ch);
}


int
laser(int side) {
#define lreturn(v) do { retval=(v); goto ret; } while (0);
    int x, y, dir;
    int ddir;
    int retval;

    /*
      dir:
        0
      3 X 1
        2
     */

    net_all_printf("LASER (%s)\n", sideinfo[side].name);

    assert(side == 0 || side == 1);
    switch (side) {
    case 0:
	x=9; y=7; dir=0;
	break;
    case 1:
	x=0; y=0; dir=2;
	break;
    }
  
    for (;;) {
	net_all_printf("pos: (%d,%d)\n", y, x);
	switch (board[x][y].piece) {
	case NONE:
	    break;
	case PHARAO:
	    lreturn(SILVER_WINS + 1-board[x][y].side); /* GAME OVER */
	case OBELISK2:
	    board[x][y].piece = OBELISK;
	    lreturn(L_PIECE);
	case OBELISK:
	    board[x][y].piece = NONE;
	    lreturn(L_PIECE);
	case DJHED:
	    ddir = (board[x][y].dir - dir) & 3;
	    dir=(dir + 1 - 2*(ddir & 1)) & 3;
	    break;
	case PYRAMID:
	    ddir = (board[x][y].dir - dir) & 3;
	    if (ddir & 2) {
		board[x][y].piece = NONE;
		lreturn(L_PIECE);
	    }
	    dir=(dir + 1 - 2*(ddir & 1)) & 3;
	    break;
	}
	switch (dir & 3) {
	case 0: y--;
	    if (y<0) lreturn(L_WALL);
	    break;
	case 1: x++;
	    if (x>BOARD_WIDTH-1) lreturn(L_WALL);
	    break;
	case 2: y++;
	    if (y>BOARD_HEIGHT-1) lreturn(L_WALL);
	    break;
	case 3: x--;
	    if (x<0) lreturn(L_WALL);
break;
	}
    }

ret:
    laser_pos[0] = x;
    laser_pos[1] = y;
    return retval;
}

int
movable(int row, int col, int side)
{
    return
	row>=0 && row<BOARD_HEIGHT &&
	col>=0 && col<BOARD_WIDTH &&
	board[col][row].piece != NONE &&
	board[col][row].side == side;
}

int
rotate(int side, int row, int col, rot_t dir)
{
    int incr;
    
    if (!movable(row, col, side))
	return E_ILLEGAL_MOVE;

    assert(dir==R_CW || dir==R_CCW);

    incr = dir==R_CW ? 1 : -1;
    board[col][row].dir = (board[col][row].dir + incr) & 3;
    return HUGE_SUCCESS;
}

int
move(int side, int row, int col, int dir, int flags) {
    if (!movable(row, col, side))
	return E_ILLEGAL_MOVE;

    /* dir =
       0 1 2
       7 X 3
       6 5 4
    */

    assert((dir>=0) || (dir<=7));
    int nrow = row - (dir<3) + ((dir>3) && (dir<7));
    int ncol = col + ((dir>1) && (dir<5)) - ((dir<1) || (dir>5));
    if ((ncol<0) || (ncol>9) || (nrow<0) || (nrow>7))
	return E_ILLEGAL_MOVE;
  
    if (flags & F_SPLIT) {
	if (board[ncol][nrow].piece || (board[col][row].piece != OBELISK2))
	    return E_ILLEGAL_MOVE;
	board[col][row].piece = OBELISK;
	board[ncol][nrow].piece = OBELISK;
	board[ncol][nrow].side = board[col][row].side;
	return HUGE_SUCCESS;
    }
  
    if (board[ncol][nrow].piece &&
	!((board[col][row].piece == DJHED) &&
	  ((board[ncol][nrow].piece != DJHED) &&
	   (board[ncol][nrow].piece != PHARAO))))
	return E_ILLEGAL_MOVE;

    struct square tmp;
    tmp = board[ncol][nrow];
    board[ncol][nrow] = board[col][row];
    board[col][row] = tmp;
  
    return HUGE_SUCCESS;
}

void
place(int row, int col, int piece, int side, int dir)
{
    if (board[col][row].piece == OBELISK &&
	board[col][row].side == side &&
	piece == OBELISK) {
	piece = OBELISK2;
    } else if (board[col][row].piece != NONE &&
	       piece != NONE) {
	fprintf(stderr, "error: trying to place a piece on a non-empty square\n");
	exit(EXIT_FAILURE);
    }

    board[col][row].piece = piece;
    board[col][row].side = side;
    board[col][row].dir = dir;
}

void
clear_board()
{
    int i, j;
    for (i = 0; i < BOARD_HEIGHT; i++)
	for (j = 0; j < BOARD_WIDTH; j++)
	    place(i, j, 0, 0, 0);
    laser_pos[0] = laser_pos[1] = -1;
}

void
setup_board_place_piece(int piece, int side, char *place_desc)
{
    int row, col, dir;

    row = place_desc[0]-'A';
    col = place_desc[1]-'0';
    dir = place_desc[2]-'0';

    if (side == 1) {
	row = BOARD_HEIGHT-row-1;
	col = BOARD_WIDTH-col-1;
	dir = (dir+2)&3;
    }

    place(row, col, piece, side, dir);
}

void
setup_board(char *desc)
{
    int side, piece, num, i;

    clear_board();

    for (side = 0; side < 2; side++) {
	i = 0;
	for (piece = 0; piece < N_PIECE_TYPES; piece++) {
	    for (num = 0; num < pieceinfo[piece].num; num++) {
		setup_board_place_piece(piece, side,
					&desc[i*3]);
		i++;
	    }
	}
    }
}

int
read_message(int side, message_t *m)
{
    char buf[MAX_MSG_LEN+1];
    char c;
    int i;
    for (i = 0; i < MAX_MSG_LEN+1; i++) {
	c = net_getc(side);
	printf("input '%c'\n", c);
	if (c == '\n')
	    break;
	buf[i] = c;
    }
    buf[i] = '\0';
    while (c != '\n') c = net_getc(side);

    switch (buf[0]) {
    case 'M': m->type = MSG_MOVE; break;
    case 'R': m->type = MSG_ROTATE; break;
    default: return E_UNKNOWN_COMMAND;
    }

    if (m->type == MSG_MOVE) {
	m->args.move.row = buf[2]-'A';
	m->args.move.col = buf[3]-'0';
	switch (buf[5]) {
	case 'N':
	    m->args.move.dir = (buf[6]=='W' ? D_NW : (buf[6]=='E' ? D_NE : D_N));
	    break;
	case 'S':
	    m->args.move.dir = (buf[6]=='W' ? D_SW : (buf[6]=='E' ? D_SE : D_S));
	    break;
	case 'E':
	    m->args.move.dir = D_E;
	    break;
	case 'W':
	    m->args.move.dir = D_W;
	    break;
	}

	net_all_printf("msg: MOVE (%d,%d) %d\n", m->args.move.row, m->args.move.col,
		       m->args.move.dir);
    } else if (m->type == MSG_ROTATE) {
	m->args.rotate.row = buf[2]-'A';
	m->args.rotate.col = buf[3]-'0';
	m->args.rotate.dir = (buf[5]=='L' ? R_CCW : R_CW);
    }

    return HUGE_SUCCESS;
}

int
execute_message(message_t msg, int side)
{
    switch (msg.type) {
    case MSG_MOVE:
	return move(side, msg.args.move.row, msg.args.move.col,
		    msg.args.move.dir, 0);
    case MSG_ROTATE:
	return rotate(side, msg.args.rotate.row, msg.args.rotate.col,
		      msg.args.rotate.dir);
    default:
	net_side_printf(side, "unknown command type %d\n", msg.type);
	return E_UNKNOWN_COMMAND;
    }
}

void
simple_game()
{
    int side = 0;

    setup_board(classic_setup);
    print_board_small();

    for (;;) {
	message_t msg;
	int r;

	net_all_printf("%s's move\n", sideinfo[side].name);
    prompt:
	net_side_printf(side, "> ");
	net_all_flush();
	
	r = read_message(side, &msg);
	switch (r) {
	case HUGE_SUCCESS: break;
	case E_UNKNOWN_COMMAND:
	    net_side_printf(side, "unknown command\n");
	    goto prompt;
	default:
	    net_side_printf(side, "uh oh (strange return value from read_message: %d)\n", r);
	    goto prompt;
	}

	r = execute_message(msg, side);
	switch (r) {
	case HUGE_SUCCESS:
	    print_board_small();
	    r = laser(side);
	    print_board_small();
	    switch (r) {
	    case SILVER_WINS:
		net_all_printf("SILVER WINS\n");
		return;
	    case RED_WINS:
		net_all_printf("RED WINS\n");
		return;
	    case L_WALL:
	    case L_PIECE:
		net_all_printf("laser hits %s at %c%c\n",
			       r==L_WALL ? "wall" : "piece",
			       laser_pos[1]+'A', laser_pos[0]+'0');
		break;
	    }
	    side = 1-side;
	    break;
	case E_ILLEGAL_MOVE:
	    net_side_printf(side, "illegal move\n");
	    goto prompt;
	default:
	    net_side_printf(side, "uh oh (strange return value from execute_message: %d)\n", r);
	    goto prompt;
	}
    }
}

int
main() {
    net_server(15239);
    simple_game();
}
